{-# LANGUAGE LambdaCase #-}
module Bank where

import Control.Concurrent

data Message a = Stop | Message {
  messageSender :: (ActorRef a),
  payload :: a
}

newtype ActorRef a = ActorRef {
  inbox :: MVar (Message a)
}

data ActorContext a = ActorContext {
  self :: ActorRef a,
  sender :: ActorRef a
}

newtype Behavior a = Behavior {
  receive :: ActorContext a -> a -> IO (Behavior a)
}

respond :: ActorContext a -> a -> IO ()
respond context msg =
  send (sender context) (self context) msg

send :: ActorRef a -> ActorRef a -> a -> IO ()
send (ActorRef recipient) sender message =
  putMVar recipient (Message sender message)

ask :: ActorRef a -> a -> IO a
ask recipient message = do
  inbox <- newEmptyMVar
  let self = ActorRef inbox
  send recipient self message
  (Message sender answer) <- takeMVar inbox
  return answer

stop :: ActorRef a -> IO ()
stop (ActorRef recipient) = putMVar recipient Stop

become :: (ActorContext a -> a -> IO (Behavior a)) -> IO (Behavior a)
become = return . Behavior

actor :: Behavior a -> IO (ActorRef a)
actor behavior = do
  inbox <- newEmptyMVar
  let self = ActorRef inbox
  let loop (Behavior behavior) = do
        msg <- takeMVar inbox
        case msg of
          Stop -> return ()
          (Message sender msg) -> do
            let context = ActorContext self sender
            newState <- behavior context msg
            loop newState
  forkIO $ loop behavior
  return self

data BankMessage =
  CreateAccount | NewBankAccount (ActorRef BankMessage) |
  Deposit Integer | Withdraw Integer |
  RequestBalance | Balance Integer |
  Transfer (ActorRef BankMessage) (ActorRef BankMessage) Integer |
  Done | Failed String

data Bank = Bank {
  accounts :: [ActorRef BankMessage]
}

bank :: IO (ActorRef BankMessage)
bank = actor (Behavior receive) where
  receive context = \case
    CreateAccount -> do
      newAccount <- account 0
      respond context $ NewBankAccount newAccount
      become receive
    Transfer from to amount -> do
      tf <- wireTransfer
      send tf (sender context) (Transfer from to amount)
      become receive

account :: Integer -> IO (ActorRef BankMessage)
account balance = actor (Behavior $ receive balance) where
  receive balance context = \case
    Deposit amount -> do
      respond context $ Done
      become $ receive (balance + amount)
    Withdraw amount ->
      if amount <= balance then do respond context $ Done
                                   become $ receive (balance - amount)
                           else do respond context $ Failed "No sufficient balance"
                                   become $ receive balance
    RequestBalance -> do
      respond context $ Balance balance
      become $ receive balance

wireTransfer :: IO (ActorRef BankMessage)
wireTransfer = actor (Behavior initialState) where
  initialState context = \case
    t@(Transfer from to amount) -> do
      send from (self context) $ Withdraw amount
      become $ waitForWithdraw (sender context) t
  waitForWithdraw client t@(Transfer from to amount) context = \case
    Done -> do
      send to (self context) $ Deposit amount
      become $ waitForDeposit client t
    Failed reason -> do
      send client (self context) $ Failed ("Withdrawal failed: " ++ reason)
      stop (self context)
      become undefined
  waitForDeposit client (Transfer from to amount) context = \case
    Done -> do
      send client (self context) $ Done
      stop (self context)
      become undefined
    Failed reason -> do
      send client (self context) $ Failed ("")
      stop (self context)
      become undefined

test :: IO ()
test = do
  theBank <- bank
  NewBankAccount account1 <- ask theBank CreateAccount
  NewBankAccount account2 <- ask theBank CreateAccount
  ask account1 $ Deposit 10000
  result <- ask theBank $ Transfer account1 account2 5000
  case result of
    Done -> putStrLn "Transfer successfull"
    Failed reason -> putStrLn reason
  Balance balance1 <- ask account1 RequestBalance
  Balance balance2 <- ask account2 RequestBalance
  putStrLn $ "Konto 1: " ++ (show balance1) ++ ", Konto 2: " ++ (show balance2)