module RoseTree where

-- Rose trees with functional updates ("zipper")

-- Rose trees:
data Tree a = Leaf a
            | Node [Tree a]
            deriving (Eq, Show)

-- A context points into a tree:
-- * either empty
-- * or a list of left subtrees, another context, and a list of right subtrees
data Ctxt a = Empty 
            | Cons [Tree a] (Ctxt a) [Tree a]

-- A tree together with a context pointing into the tree
newtype Loc a = Loc (Tree a, Ctxt a) 

go_left :: Loc a-> Loc a
go_left (Loc(t, c)) = case c of
 Cons (l:le) up ri -> Loc(l, Cons le up (t:ri))
 _                 -> error "go_left of first"

go_right :: Loc a-> Loc a
go_right (Loc(t, c)) = case c of
 Cons le up (r:ri) -> Loc(r, Cons (t:le) up ri)
 _                 -> error "go_right of last"

go_up :: Loc a-> Loc a
go_up (Loc (t, c)) = case c of
  Empty -> error "go_up of empty"
  Cons le up ri -> 
    Loc (Node (reverse le ++  t:ri), up) 

go_down :: Loc a-> Loc a
go_down (Loc (t, c)) = case t of
  Leaf _ -> error "go_down at leaf"
  Node [] -> error "go_down at empty"
  Node (t:ts) -> Loc (t, Cons [] c ts)

insert_left :: Tree a-> Loc a-> Loc a
insert_left t1 (Loc (t, c)) = case c of
  Empty -> error "insert_left: insert at empty"
  Cons le up ri -> Loc(t, Cons (t1:le) up ri) 

insert_right :: Tree a-> Loc a-> Loc a
insert_right t1 (Loc (t, c)) = case c of
  Empty -> error "insert_right: insert at empty"
  Cons le up ri -> Loc(t, Cons le up (t1:ri))

insert_down :: Tree a-> Loc a-> Loc a
insert_down t1 (Loc(t, c)) = case t of
  Leaf _ -> error "insert_down: insert at leaf"
  Node ts -> Loc(t1, Cons [] c ts)

update :: Tree a-> Loc a-> Loc a
update t (Loc (_, c)) = Loc (t, c)

delete :: Loc a-> Loc a
delete (Loc(_, p)) = case p of
  Empty -> Loc(Node [], Empty)
  Cons le up (r:ri) -> Loc(r, Cons le up ri)
  Cons (l:le) up [] -> Loc(l, Cons le up [])
  Cons [] up [] -> Loc(Node [], up)

-- Helper function (creates a focus)
top :: Tree a-> Loc a
top t = (Loc (t, Empty))  

-- Navigation by path (cf. the simple rose trees)
path :: Loc a-> [Int]-> Loc a
path l [] = l
path l (i:ps) 
  | i == 0 = path (go_down l) ps
  | i > 0  = path (go_left l) (i-1: ps)


-- Example tree
t = Node [ Leaf "-"
         , Node [Leaf "*", Leaf "a", Leaf "b"]
         , Node [Leaf "*", Leaf "c", Leaf "d"]
         ]

-- For debugging:
instance Show a=> Show(Loc a) where
 show (Loc (t, Empty)) = show t
 show (Loc (t, Cons ls c ts)) =
   replicate (length ls) '<' ++ show t++ replicate (length ts) '>'
